@@ -21,14 +21,100 @@
         .value = AWS_BYTE_CUR_INIT_FROM_STRING_LITERAL(VALUE),                                                         \
     }
 
-static int s_validate_mpu_mock_server_metrics(struct aws_array_list *metrics_list) {
-    /* Check the size of the metrics should be the same as the number of requests, which should be create MPU, two
-     * upload parts and one complete MPU */
-    ASSERT_UINT_EQUALS(4, aws_array_list_length(metrics_list));
-    struct aws_s3_request_metrics *metrics = NULL;
+static int s_validate_time_metrics(struct aws_s3_request_metrics *metrics, bool is_last_attempt) {
+    uint64_t start = 0, end = 0, duration = 0;
+    int error_code = aws_s3_request_metrics_get_error_code(metrics);
+    bool is_success = (error_code == AWS_ERROR_SUCCESS);
 
-    /* First metrics should be the CreateMPU */
-    aws_array_list_get_at(metrics_list, (void **)&metrics, 0);
+    /* Always available */
+    aws_s3_request_metrics_get_s3_request_first_attempt_start_timestamp_ns(metrics, &start);
+    ASSERT_TRUE(start > 0);
+    /* Available on last attempt only */
+    if (is_last_attempt) {
+        ASSERT_SUCCESS(aws_s3_request_metrics_get_s3_request_last_attempt_end_timestamp_ns(metrics, &end));
+        ASSERT_TRUE(end > 0);
+        ASSERT_SUCCESS(aws_s3_request_metrics_get_s3_request_total_duration_ns(metrics, &duration));
+        ASSERT_TRUE(duration > 0);
+        ASSERT_UINT_EQUALS(end - start, duration);
+    } else {
+        ASSERT_FAILS(aws_s3_request_metrics_get_s3_request_last_attempt_end_timestamp_ns(metrics, &end));
+        ASSERT_FAILS(aws_s3_request_metrics_get_s3_request_total_duration_ns(metrics, &end));
+    }
+
+    aws_s3_request_metrics_get_start_timestamp_ns(metrics, &start);
+    ASSERT_TRUE(start > 0);
+    aws_s3_request_metrics_get_end_timestamp_ns(metrics, &end);
+    ASSERT_TRUE(end > 0);
+    aws_s3_request_metrics_get_total_duration_ns(metrics, &duration);
+    ASSERT_TRUE(duration > 0);
+    ASSERT_UINT_EQUALS(end - start, duration);
+
+    /* Available on last attempt with success only */
+    if (is_last_attempt && is_success) {
+        ASSERT_SUCCESS(aws_s3_request_metrics_get_send_start_timestamp_ns(metrics, &start));
+        ASSERT_TRUE(start > 0);
+        ASSERT_SUCCESS(aws_s3_request_metrics_get_send_end_timestamp_ns(metrics, &end));
+        ASSERT_TRUE(end > 0);
+        ASSERT_SUCCESS(aws_s3_request_metrics_get_sending_duration_ns(metrics, &duration));
+        ASSERT_TRUE(duration > 0);
+        ASSERT_UINT_EQUALS(end - start, duration);
+
+        ASSERT_SUCCESS(aws_s3_request_metrics_get_receive_start_timestamp_ns(metrics, &start));
+        ASSERT_TRUE(start > 0);
+        ASSERT_SUCCESS(aws_s3_request_metrics_get_receive_end_timestamp_ns(metrics, &end));
+        ASSERT_TRUE(end > 0);
+        ASSERT_SUCCESS(aws_s3_request_metrics_get_receiving_duration_ns(metrics, &duration));
+        ASSERT_TRUE(duration > 0);
+        ASSERT_UINT_EQUALS(end - start, duration);
+
+        ASSERT_SUCCESS(aws_s3_request_metrics_get_service_call_duration_ns(metrics, &duration));
+        ASSERT_TRUE(duration > 0);
+
+        ASSERT_SUCCESS(aws_s3_request_metrics_get_conn_acquire_start_timestamp_ns(metrics, &start));
+        ASSERT_TRUE(start > 0);
+        ASSERT_SUCCESS(aws_s3_request_metrics_get_conn_acquire_end_timestamp_ns(metrics, &end));
+        ASSERT_TRUE(end > 0);
+        ASSERT_SUCCESS(aws_s3_request_metrics_get_conn_acquire_duration_ns(metrics, &duration));
+        ASSERT_TRUE(duration > 0);
+        ASSERT_UINT_EQUALS(end - start, duration);
+
+        ASSERT_SUCCESS(aws_s3_request_metrics_get_sign_start_timestamp_ns(metrics, &start));
+        ASSERT_TRUE(start > 0);
+        ASSERT_SUCCESS(aws_s3_request_metrics_get_sign_end_timestamp_ns(metrics, &end));
+        ASSERT_TRUE(end > 0);
+        ASSERT_SUCCESS(aws_s3_request_metrics_get_signing_duration_ns(metrics, &duration));
+        ASSERT_TRUE(duration > 0);
+        ASSERT_UINT_EQUALS(end - start, duration);
+
+        if (metrics->req_resp_info_metrics.request_type == AWS_S3_REQUEST_TYPE_GET_OBJECT) {
+            ASSERT_SUCCESS(aws_s3_request_metrics_get_delivery_start_timestamp_ns(metrics, &start));
+            ASSERT_TRUE(start > 0);
+            ASSERT_SUCCESS(aws_s3_request_metrics_get_delivery_end_timestamp_ns(metrics, &end));
+            ASSERT_TRUE(end > 0);
+            ASSERT_SUCCESS(aws_s3_request_metrics_get_delivery_duration_ns(metrics, &duration));
+            ASSERT_TRUE(duration > 0);
+            ASSERT_UINT_EQUALS(end - start, duration);
+        }
+    }
+
+    if (metrics->crt_info_metrics.retry_attempt > 0) {
+        ASSERT_SUCCESS(aws_s3_request_metrics_get_retry_delay_start_timestamp_ns(metrics, &start));
+        ASSERT_TRUE(start > 0);
+        ASSERT_SUCCESS(aws_s3_request_metrics_get_retry_delay_end_timestamp_ns(metrics, &end));
+        ASSERT_TRUE(end > 0);
+        ASSERT_SUCCESS(aws_s3_request_metrics_get_retry_delay_duration_ns(metrics, &duration));
+        ASSERT_TRUE(duration > 0);
+        ASSERT_UINT_EQUALS(end - start, duration);
+    } else {
+        ASSERT_FAILS(aws_s3_request_metrics_get_retry_delay_start_timestamp_ns(metrics, &start));
+        ASSERT_FAILS(aws_s3_request_metrics_get_retry_delay_end_timestamp_ns(metrics, &end));
+        ASSERT_FAILS(aws_s3_request_metrics_get_retry_delay_duration_ns(metrics, &duration));
+    }
+
+    return AWS_OP_SUCCESS;
+}
+
+static int s_validate_create_multipart_upload_metrics(struct aws_s3_request_metrics *metrics) {
     struct aws_http_headers *response_headers = NULL;
     ASSERT_SUCCESS(aws_s3_request_metrics_get_response_headers(metrics, &response_headers));
     const struct aws_string *request_id = NULL;
@@ -52,38 +138,13 @@ static int s_validate_mpu_mock_server_metrics(struct aws_array_list *metrics_lis
     ASSERT_TRUE(host_address->len > 0);
     aws_thread_id_t thread_id = 0;
     ASSERT_SUCCESS(aws_s3_request_metrics_get_thread_id(metrics, &thread_id));
-    size_t connection_id = 0;
-    ASSERT_SUCCESS(aws_s3_request_metrics_get_connection_id(metrics, &connection_id));
+    size_t connection_ptr = 0;
+    ASSERT_SUCCESS(aws_s3_request_metrics_get_connection_id(metrics, &connection_ptr));
     ASSERT_UINT_EQUALS(AWS_ERROR_SUCCESS, aws_s3_request_metrics_get_error_code(metrics));
+
     /* Get all those time stamp */
-    uint64_t time_stamp = 0;
-    aws_s3_request_metrics_get_start_timestamp_ns(metrics, &time_stamp);
-    ASSERT_FALSE(time_stamp == 0);
-    time_stamp = 0;
-    aws_s3_request_metrics_get_end_timestamp_ns(metrics, &time_stamp);
-    ASSERT_FALSE(time_stamp == 0);
-    time_stamp = 0;
-    aws_s3_request_metrics_get_total_duration_ns(metrics, &time_stamp);
-    ASSERT_FALSE(time_stamp == 0);
-    time_stamp = 0;
-    ASSERT_SUCCESS(aws_s3_request_metrics_get_send_start_timestamp_ns(metrics, &time_stamp));
-    ASSERT_FALSE(time_stamp == 0);
-    time_stamp = 0;
-    ASSERT_SUCCESS(aws_s3_request_metrics_get_send_end_timestamp_ns(metrics, &time_stamp));
-    ASSERT_FALSE(time_stamp == 0);
-    time_stamp = 0;
-    ASSERT_SUCCESS(aws_s3_request_metrics_get_sending_duration_ns(metrics, &time_stamp));
-    ASSERT_FALSE(time_stamp == 0);
-    time_stamp = 0;
-    ASSERT_SUCCESS(aws_s3_request_metrics_get_receive_start_timestamp_ns(metrics, &time_stamp));
-    ASSERT_FALSE(time_stamp == 0);
-    time_stamp = 0;
-    ASSERT_SUCCESS(aws_s3_request_metrics_get_receive_end_timestamp_ns(metrics, &time_stamp));
-    ASSERT_FALSE(time_stamp == 0);
-    time_stamp = 0;
-    ASSERT_SUCCESS(aws_s3_request_metrics_get_receiving_duration_ns(metrics, &time_stamp));
-    ASSERT_FALSE(time_stamp == 0);
-    time_stamp = 0;
+    ASSERT_SUCCESS(s_validate_time_metrics(metrics, true));
+
     enum aws_s3_request_type request_type = 0;
     aws_s3_request_metrics_get_request_type(metrics, &request_type);
     ASSERT_UINT_EQUALS(AWS_S3_REQUEST_TYPE_CREATE_MULTIPART_UPLOAD, request_type);
@@ -92,38 +153,171 @@ static int s_validate_mpu_mock_server_metrics(struct aws_array_list *metrics_lis
     ASSERT_SUCCESS(aws_s3_request_metrics_get_operation_name(metrics, &operation_name));
     ASSERT_STR_EQUALS("CreateMultipartUpload", aws_string_c_str(operation_name));
 
-    /* Second metrics should be the Upload Part */
-    aws_array_list_get_at(metrics_list, (void **)&metrics, 1);
+    return AWS_OP_SUCCESS;
+}
+
+static int s_validate_upload_part_metrics(struct aws_s3_request_metrics *metrics, bool is_last_attempt) {
+    struct aws_http_headers *response_headers = NULL;
     struct aws_byte_cursor header_value;
+    enum aws_s3_request_type request_type = 0;
+    const struct aws_string *operation_name = NULL;
+
     AWS_ZERO_STRUCT(header_value);
     response_headers = NULL;
-    ASSERT_SUCCESS(aws_s3_request_metrics_get_response_headers(metrics, &response_headers));
-    ASSERT_SUCCESS(aws_http_headers_get(response_headers, aws_byte_cursor_from_c_str("ETag"), &header_value));
-    ASSERT_TRUE(aws_byte_cursor_eq_c_str(&header_value, "b54357faf0632cce46e942fa68356b38"));
-    ASSERT_SUCCESS(aws_http_headers_get(response_headers, aws_byte_cursor_from_c_str("Connection"), &header_value));
-    ASSERT_TRUE(aws_byte_cursor_eq_c_str(&header_value, "keep-alive"));
+    if (metrics->req_resp_info_metrics.response_status != -1) {
+        ASSERT_SUCCESS(aws_s3_request_metrics_get_response_headers(metrics, &response_headers));
+        ASSERT_SUCCESS(aws_http_headers_get(response_headers, aws_byte_cursor_from_c_str("ETag"), &header_value));
+        ASSERT_TRUE(aws_byte_cursor_eq_c_str(&header_value, "b54357faf0632cce46e942fa68356b38"));
+        ASSERT_SUCCESS(aws_http_headers_get(response_headers, aws_byte_cursor_from_c_str("Connection"), &header_value));
+        ASSERT_TRUE(aws_byte_cursor_eq_c_str(&header_value, "keep-alive"));
+    }
+
     request_type = 0;
     aws_s3_request_metrics_get_request_type(metrics, &request_type);
     ASSERT_UINT_EQUALS(AWS_S3_REQUEST_TYPE_UPLOAD_PART, request_type);
     ASSERT_SUCCESS(aws_s3_request_metrics_get_operation_name(metrics, &operation_name));
     ASSERT_STR_EQUALS("UploadPart", aws_string_c_str(operation_name));
 
-    /* Third metrics still be Upload Part */
-    aws_array_list_get_at(metrics_list, (void **)&metrics, 2);
-    request_type = 0;
-    aws_s3_request_metrics_get_request_type(metrics, &request_type);
-    ASSERT_UINT_EQUALS(AWS_S3_REQUEST_TYPE_UPLOAD_PART, request_type);
-    ASSERT_SUCCESS(aws_s3_request_metrics_get_operation_name(metrics, &operation_name));
-    ASSERT_STR_EQUALS("UploadPart", aws_string_c_str(operation_name));
+    ASSERT_SUCCESS(s_validate_time_metrics(metrics, is_last_attempt));
+
+    return AWS_OP_SUCCESS;
+}
+
+static int s_validate_complete_multipart_upload_metrics(struct aws_s3_request_metrics *metrics) {
+    enum aws_s3_request_type request_type = 0;
+    const struct aws_string *operation_name = NULL;
 
-    /* Fourth should be complete MPU */
-    aws_array_list_get_at(metrics_list, (void **)&metrics, 3);
-    request_type = 0;
     aws_s3_request_metrics_get_request_type(metrics, &request_type);
     ASSERT_UINT_EQUALS(AWS_S3_REQUEST_TYPE_COMPLETE_MULTIPART_UPLOAD, request_type);
     ASSERT_SUCCESS(aws_s3_request_metrics_get_operation_name(metrics, &operation_name));
     ASSERT_STR_EQUALS("CompleteMultipartUpload", aws_string_c_str(operation_name));
-    /* All the rest should be similar */
+
+    ASSERT_SUCCESS(s_validate_time_metrics(metrics, true));
+
+    return AWS_OP_SUCCESS;
+}
+
+static int s_validate_abort_multipart_upload_metrics(struct aws_s3_request_metrics *metrics) {
+    enum aws_s3_request_type request_type = 0;
+    const struct aws_string *operation_name = NULL;
+
+    aws_s3_request_metrics_get_request_type(metrics, &request_type);
+    ASSERT_UINT_EQUALS(AWS_S3_REQUEST_TYPE_ABORT_MULTIPART_UPLOAD, request_type);
+    ASSERT_SUCCESS(aws_s3_request_metrics_get_operation_name(metrics, &operation_name));
+    ASSERT_STR_EQUALS("AbortMultipartUpload", aws_string_c_str(operation_name));
+
+    ASSERT_SUCCESS(s_validate_time_metrics(metrics, true));
+
+    return AWS_OP_SUCCESS;
+}
+
+static int s_validate_mpu_mock_server_metrics(struct aws_array_list *metrics_list, uint32_t expected_length) {
+    /* Check the size of the metrics should be the same as the number of requests, which should be create MPU, two
+     * upload parts and one complete MPU */
+    ASSERT_UINT_EQUALS(expected_length, aws_array_list_length(metrics_list));
+    struct aws_s3_request_metrics *metrics = NULL;
+
+    /* First metrics should be the CreateMPU */
+    aws_array_list_get_at(metrics_list, (void **)&metrics, 0);
+    ASSERT_SUCCESS(s_validate_create_multipart_upload_metrics(metrics));
+
+    /* All of the middle should be Upload Parts*/
+    for (size_t i = 1; i < aws_array_list_length(metrics_list) - 1; i++) {
+        metrics = NULL;
+        aws_array_list_get_at(metrics_list, (void **)&metrics, i);
+        ASSERT_SUCCESS(s_validate_upload_part_metrics(metrics, true)); /* assuming all requests were success */
+    }
+
+    /* Last metrics should be CompleteMPU*/
+    metrics = NULL;
+    aws_array_list_get_at(metrics_list, (void **)&metrics, aws_array_list_length(metrics_list) - 1);
+    ASSERT_SUCCESS(s_validate_complete_multipart_upload_metrics(metrics));
+
+    return AWS_OP_SUCCESS;
+}
+
+static int s_validate_retry_metrics(struct aws_array_list *metrics_list, uint32_t expected_failures, uint32_t parts) {
+    struct aws_s3_request_metrics *metrics = NULL, *metrics2 = NULL;
+
+    /* First metrics should be the CreateMPU */
+    aws_array_list_get_at(metrics_list, (void **)&metrics, 0);
+    ASSERT_SUCCESS(s_validate_create_multipart_upload_metrics(metrics));
+
+    /* All of the middle should be Upload Parts*/
+    /* This check assumes each request fails 'expected_failures' number of times and then succeeds.
+     * So for each part, we would have expected_failures + 1 attempts.
+     * We make sure all of the attempts have the same request_first_attempt_start_time and the last attempt has an end
+     * time. */
+    size_t failed_count = 0;
+    for (size_t i = 1; i < aws_array_list_length(metrics_list) - 1; i = i + expected_failures + 1) {
+        aws_array_list_get_at(metrics_list, &metrics, i);
+        for (size_t j = i; j < i + expected_failures; j++) {
+            aws_array_list_get_at(metrics_list, &metrics2, j + 1);
+            ASSERT_TRUE(metrics->crt_info_metrics.error_code != AWS_ERROR_SUCCESS);
+            failed_count++;
+            aws_array_list_get_at(metrics_list, (void **)&metrics, j);
+            ASSERT_SUCCESS(s_validate_upload_part_metrics(metrics, false));
+            ASSERT_INT_EQUALS(
+                metrics->time_metrics.s3_request_first_attempt_start_timestamp_ns,
+                metrics2->time_metrics.s3_request_first_attempt_start_timestamp_ns);
+            ASSERT_INT_EQUALS(metrics->crt_info_metrics.retry_attempt + 1, metrics2->crt_info_metrics.retry_attempt);
+            ASSERT_TRUE(metrics->crt_info_metrics.request_ptr == metrics2->crt_info_metrics.request_ptr);
+            metrics = metrics2;
+        }
+        ASSERT_SUCCESS(s_validate_upload_part_metrics(metrics, true));
+    }
+
+    ASSERT_UINT_EQUALS(expected_failures * parts, failed_count);
+
+    /* Last metrics should be CompleteMPU*/
+    metrics = NULL;
+    aws_array_list_get_at(metrics_list, (void **)&metrics, aws_array_list_length(metrics_list) - 1);
+    ASSERT_SUCCESS(s_validate_complete_multipart_upload_metrics(metrics));
+
+    return AWS_OP_SUCCESS;
+}
+
+static int s_validate_fail_metrics(struct aws_array_list *metrics_list, uint32_t parts) {
+    struct aws_s3_request_metrics *metrics = NULL, *metrics2 = NULL;
+
+    /* First metrics should be the CreateMPU */
+    aws_array_list_get_at(metrics_list, (void **)&metrics, 0);
+    ASSERT_SUCCESS(s_validate_create_multipart_upload_metrics(metrics));
+
+    /* It is difficult to simulate forced failure and be precise about how a request fails if there are multiple
+     * connections. Assuming, the first request itself is failing, all of the other parts are force cancelled. If there
+     * are n parts, we would record n + 5 metrics. 5 retries for the first part alone. */
+
+    /* First part fails 5 times */
+    aws_array_list_get_at(metrics_list, &metrics, 1);
+    ASSERT_TRUE(metrics->crt_info_metrics.error_code != AWS_ERROR_SUCCESS);
+    size_t request_ptr = (size_t)metrics->crt_info_metrics.request_ptr;
+    int64_t request_start_time = metrics->time_metrics.s3_request_first_attempt_start_timestamp_ns;
+    for (size_t i = 1; i < 6; i++) {
+        aws_array_list_get_at(metrics_list, &metrics2, i + 1);
+        ASSERT_INT_EQUALS(
+            metrics->time_metrics.s3_request_first_attempt_start_timestamp_ns,
+            metrics2->time_metrics.s3_request_first_attempt_start_timestamp_ns);
+        ASSERT_INT_EQUALS(metrics->crt_info_metrics.retry_attempt + 1, metrics2->crt_info_metrics.retry_attempt);
+        ASSERT_TRUE(metrics2->crt_info_metrics.error_code != AWS_ERROR_SUCCESS);
+        ASSERT_SUCCESS(s_validate_upload_part_metrics(metrics, false));
+        ASSERT_INT_EQUALS(request_start_time, metrics2->time_metrics.s3_request_first_attempt_start_timestamp_ns);
+        ASSERT_TRUE(request_ptr == (size_t)metrics2->crt_info_metrics.request_ptr);
+        metrics = metrics2;
+    }
+    ASSERT_SUCCESS(s_validate_upload_part_metrics(metrics, true));
+
+    /* Rest of the request should have been cancelled */
+    for (size_t i = 7; i < parts + 6; i++) {
+        aws_array_list_get_at(metrics_list, &metrics, i);
+        ASSERT_TRUE(metrics->crt_info_metrics.error_code == AWS_ERROR_S3_CANCELED);
+        ASSERT_SUCCESS(s_validate_upload_part_metrics(metrics, true));
+    }
+
+    /* Last metrics should be AbortMPU*/
+    metrics = NULL;
+    aws_array_list_get_at(metrics_list, (void **)&metrics, aws_array_list_length(metrics_list) - 1);
+    ASSERT_SUCCESS(s_validate_abort_multipart_upload_metrics(metrics));
 
     return AWS_OP_SUCCESS;
 }
@@ -159,7 +353,8 @@ TEST_CASE(multipart_upload_mock_server) {
     struct aws_s3_meta_request_test_results out_results;
     aws_s3_meta_request_test_results_init(&out_results, allocator);
     ASSERT_SUCCESS(aws_s3_tester_send_meta_request_with_options(&tester, &put_options, &out_results));
-    ASSERT_SUCCESS(s_validate_mpu_mock_server_metrics(&out_results.synced_data.metrics));
+    ASSERT_SUCCESS(
+        s_validate_mpu_mock_server_metrics(&out_results.synced_data.metrics, 4 /*1 create, 1 complete, 2 parts*/));
     aws_s3_meta_request_test_results_clean_up(&out_results);
     aws_s3_client_release(client);
     aws_s3_tester_clean_up(&tester);
@@ -167,6 +362,147 @@ TEST_CASE(multipart_upload_mock_server) {
     return AWS_OP_SUCCESS;
 }
 
+static void s_upload_part_with_n_retries(struct aws_s3_request *request, struct aws_http_message *message) {
+    if (message == NULL) {
+        return;
+    }
+
+    struct aws_s3_meta_request *meta_request = request->meta_request;
+    struct aws_s3_client *client = meta_request->client;
+    struct aws_s3_tester *tester = client->shutdown_callback_user_data; // or similar
+    uint32_t n = (uint32_t)(uintptr_t)tester->user_data;
+
+    if (request->num_times_prepared < n - 1) {
+        struct aws_http_header throttle_header = {
+            .name = aws_byte_cursor_from_c_str("force_throttle"),
+            .value = aws_byte_cursor_from_c_str("true"),
+        };
+        aws_http_message_add_header(message, throttle_header);
+    }
+}
+
+TEST_CASE(multipart_upload_with_n_retries_mock_server) {
+    (void)ctx;
+
+    struct aws_s3_tester tester;
+    ASSERT_SUCCESS(aws_s3_tester_init(allocator, &tester));
+    tester.user_data = (void *)(uintptr_t)2; /* Fail 3 (n-1) times, succeed on 4th (nth) */
+    int part_size = 5;
+
+    struct aws_s3_tester_client_options client_options = {
+        .part_size = MB_TO_BYTES(part_size),
+        .tls_usage = AWS_S3_TLS_DISABLED,
+        .max_active_connections_override = 1,
+    };
+
+    struct aws_s3_client *client = NULL;
+    ASSERT_SUCCESS(aws_s3_tester_client_new(&tester, &client_options, &client));
+    struct aws_s3_client_vtable *patched_client_vtable = aws_s3_tester_patch_client_vtable(&tester, client, NULL);
+    patched_client_vtable->after_prepare_upload_part_finish_stub = s_upload_part_with_n_retries;
+
+    int object_size = 10;
+    int parts = object_size / part_size;
+
+    struct aws_byte_cursor object_path = aws_byte_cursor_from_c_str("/default");
+    {
+        /* 1. Trailer checksum */
+        struct aws_s3_tester_meta_request_options put_options = {
+            .allocator = allocator,
+            .meta_request_type = AWS_S3_META_REQUEST_TYPE_PUT_OBJECT,
+            .client = client,
+            .checksum_algorithm = AWS_SCA_CRC32,
+            .validate_get_response_checksum = false,
+            .put_options =
+                {
+                    .object_size_mb = object_size,
+                    .object_path_override = object_path,
+                },
+            .mock_server = true,
+        };
+
+        struct aws_s3_meta_request_test_results meta_request_test_results;
+
+        // check if number of metrics received for each part is n
+        aws_s3_meta_request_test_results_init(&meta_request_test_results, allocator);
+        ASSERT_SUCCESS(aws_s3_tester_send_meta_request_with_options(&tester, &put_options, &meta_request_test_results));
+        uint32_t expected_failures = (uint32_t)(uintptr_t)tester.user_data - 1;
+        ASSERT_SUCCESS(
+            s_validate_retry_metrics(&meta_request_test_results.synced_data.metrics, expected_failures, parts));
+
+        aws_s3_meta_request_test_results_clean_up(&meta_request_test_results);
+    }
+    aws_s3_client_release(client);
+    aws_s3_tester_clean_up(&tester);
+
+    return AWS_OP_SUCCESS;
+}
+
+static void s_upload_part_force_fail(struct aws_s3_request *request, struct aws_http_message *message) {
+    (void)request;
+    if (message == NULL) {
+        return;
+    }
+
+    struct aws_http_header throttle_header = {
+        .name = aws_byte_cursor_from_c_str("force_throttle"),
+        .value = aws_byte_cursor_from_c_str("true"),
+    };
+    aws_http_message_add_header(message, throttle_header);
+}
+
+TEST_CASE(multipart_upload_failure_with_mock_server) {
+    (void)ctx;
+
+    struct aws_s3_tester tester;
+    ASSERT_SUCCESS(aws_s3_tester_init(allocator, &tester));
+    int part_size = 5;
+
+    struct aws_s3_tester_client_options client_options = {
+        .part_size = MB_TO_BYTES(part_size),
+        .tls_usage = AWS_S3_TLS_DISABLED,
+        .max_active_connections_override = 1,
+    };
+
+    struct aws_s3_client *client = NULL;
+    ASSERT_SUCCESS(aws_s3_tester_client_new(&tester, &client_options, &client));
+    struct aws_s3_client_vtable *patched_client_vtable = aws_s3_tester_patch_client_vtable(&tester, client, NULL);
+    patched_client_vtable->after_prepare_upload_part_finish_stub = s_upload_part_force_fail;
+
+    int object_size = 10;
+    int parts = object_size / part_size;
+
+    struct aws_byte_cursor object_path = aws_byte_cursor_from_c_str("/default");
+    {
+        struct aws_s3_tester_meta_request_options put_options = {
+            .allocator = allocator,
+            .meta_request_type = AWS_S3_META_REQUEST_TYPE_PUT_OBJECT,
+            .client = client,
+            .checksum_algorithm = AWS_SCA_CRC32,
+            .validate_get_response_checksum = false,
+            .put_options =
+                {
+                    .object_size_mb = object_size,
+                    .object_path_override = object_path,
+                },
+            .mock_server = true,
+            .validate_type = AWS_S3_TESTER_VALIDATE_TYPE_NO_VALIDATE,
+        };
+
+        struct aws_s3_meta_request_test_results meta_request_test_results;
+
+        // check if number of metrics received for each part is n
+        aws_s3_meta_request_test_results_init(&meta_request_test_results, allocator);
+        ASSERT_SUCCESS(aws_s3_tester_send_meta_request_with_options(&tester, &put_options, &meta_request_test_results));
+        ASSERT_SUCCESS(s_validate_fail_metrics(&meta_request_test_results.synced_data.metrics, parts));
+
+        aws_s3_meta_request_test_results_clean_up(&meta_request_test_results);
+    }
+    aws_s3_client_release(client);
+    aws_s3_tester_clean_up(&tester);
+
+    return AWS_OP_SUCCESS;
+}
+
 /* Singleton used by tests in this file */
 static struct get_requests_header_tester {
     struct aws_allocator *alloc;
@@ -251,7 +587,8 @@ TEST_CASE(multipart_upload_unsigned_with_trailer_checksum_mock_server) {
     struct aws_s3_meta_request_test_results out_results;
     aws_s3_meta_request_test_results_init(&out_results, allocator);
     ASSERT_SUCCESS(aws_s3_tester_send_meta_request_with_options(&tester, &put_options, &out_results));
-    ASSERT_SUCCESS(s_validate_mpu_mock_server_metrics(&out_results.synced_data.metrics));
+    ASSERT_SUCCESS(
+        s_validate_mpu_mock_server_metrics(&out_results.synced_data.metrics, 4 /*1 create, 1 complete, 2 parts*/));
 
     /**
      * Check the recorded headers.
@@ -435,13 +772,99 @@ TEST_CASE(multipart_upload_with_network_interface_names_mock_server) {
 }
 
 /* Total hack to flip the bytes. */
-static void s_after_prepare_upload_part_finish(struct aws_s3_request *request) {
-    if (request->num_times_prepared > 1) {
+static void s_after_prepare_upload_part_finish_stub(struct aws_s3_request *request, struct aws_http_message *message) {
+    (void)message;
+    if (request->num_times_prepared == 0 && message != NULL) {
+        struct aws_http_header throttle_header = {
+            .name = aws_byte_cursor_from_c_str("force_throttle"),
+            .value = aws_byte_cursor_from_c_str("true"),
+        };
+        aws_http_message_add_header(message, throttle_header);
+    }
+    if (request->num_times_prepared > 0) {
         /* mock that the body buffer was messed up in memory */
         request->request_body.buffer[1]++;
     }
 }
 
+static void s_after_prepare_upload_part_finish_stub_retry_before_finish_sending(
+    struct aws_s3_request *request,
+    struct aws_http_message *message) {
+    if (request->num_times_prepared == 0 && message != NULL) {
+        struct aws_http_header before_finish_header = {
+            .name = aws_byte_cursor_from_c_str("before_finish"),
+            .value = aws_byte_cursor_from_c_str("true"),
+        };
+        aws_http_message_add_header(message, before_finish_header);
+        struct aws_http_header throttle_header = {
+            .name = aws_byte_cursor_from_c_str("force_throttle"),
+            .value = aws_byte_cursor_from_c_str("true"),
+        };
+        aws_http_message_add_header(message, throttle_header);
+    }
+    if (request->num_times_prepared > 0 && request->request_body.buffer != NULL) {
+        /* mock that the body buffer was messed up in memory */
+        request->request_body.buffer[1]++;
+    }
+}
+
+/**
+ * This test is built for
+ * 1. The retry happens before the upload has finished.
+ */
+TEST_CASE(multipart_upload_checksum_with_retry_before_finish_mock_server) {
+    (void)ctx;
+    struct aws_s3_tester tester;
+    ASSERT_SUCCESS(aws_s3_tester_init(allocator, &tester));
+    struct aws_s3_tester_client_options client_options = {
+        .part_size = MB_TO_BYTES(5),
+        .tls_usage = AWS_S3_TLS_DISABLED,
+    };
+
+    struct aws_s3_client *client = NULL;
+    ASSERT_SUCCESS(aws_s3_tester_client_new(&tester, &client_options, &client));
+    struct aws_s3_client_vtable *patched_client_vtable = aws_s3_tester_patch_client_vtable(&tester, client, NULL);
+    patched_client_vtable->after_prepare_upload_part_finish_stub =
+        s_after_prepare_upload_part_finish_stub_retry_before_finish_sending;
+
+    struct aws_byte_cursor object_path = aws_byte_cursor_from_c_str("/default");
+    {
+        /* 1. Trailer checksum */
+        struct aws_s3_tester_meta_request_options put_options = {
+            .allocator = allocator,
+            .meta_request_type = AWS_S3_META_REQUEST_TYPE_PUT_OBJECT,
+            .client = client,
+            .checksum_algorithm = AWS_SCA_CRC32,
+            .validate_get_response_checksum = false,
+            .put_options =
+                {
+                    .object_size_mb = 10,
+                    .object_path_override = object_path,
+                },
+            .mock_server = true,
+        };
+
+        struct aws_s3_meta_request_test_results meta_request_test_results;
+        aws_s3_meta_request_test_results_init(&meta_request_test_results, allocator);
+
+        ASSERT_SUCCESS(aws_s3_tester_send_meta_request_with_options(&tester, &put_options, &meta_request_test_results));
+
+        ASSERT_INT_EQUALS(meta_request_test_results.upload_review.part_count, 2);
+        /* Note: the data we currently generate is always the same,
+         * The retry got the messed up data, while the first run never actually finish reading the bytes, so the messed
+         * up data checksum got to be sent. */
+        ASSERT_STR_EQUALS(
+            "dKYRxA==", aws_string_c_str(meta_request_test_results.upload_review.part_checksums_array[0]));
+        ASSERT_STR_EQUALS(
+            "dxV2Sw==", aws_string_c_str(meta_request_test_results.upload_review.part_checksums_array[1]));
+        aws_s3_meta_request_test_results_clean_up(&meta_request_test_results);
+    }
+    aws_s3_client_release(client);
+    aws_s3_tester_clean_up(&tester);
+
+    return AWS_OP_SUCCESS;
+}
+
 /**
  * This test is built for
  * 1. We had a memory leak when the retry was triggered and the checksum was calculated.
@@ -461,9 +884,9 @@ TEST_CASE(multipart_upload_checksum_with_retry_mock_server) {
     struct aws_s3_client *client = NULL;
     ASSERT_SUCCESS(aws_s3_tester_client_new(&tester, &client_options, &client));
     struct aws_s3_client_vtable *patched_client_vtable = aws_s3_tester_patch_client_vtable(&tester, client, NULL);
-    patched_client_vtable->after_prepare_upload_part_finish = s_after_prepare_upload_part_finish;
+    patched_client_vtable->after_prepare_upload_part_finish_stub = s_after_prepare_upload_part_finish_stub;
 
-    struct aws_byte_cursor object_path = aws_byte_cursor_from_c_str("/throttle");
+    struct aws_byte_cursor object_path = aws_byte_cursor_from_c_str("/default");
     {
         /* 1. Trailer checksum */
         struct aws_s3_tester_meta_request_options put_options = {
@@ -531,6 +954,60 @@ TEST_CASE(multipart_upload_checksum_with_retry_mock_server) {
     return AWS_OP_SUCCESS;
 }
 
+TEST_CASE(multipart_upload_checksum_fio_with_retry_mock_server) {
+    (void)ctx;
+    struct aws_s3_tester tester;
+    ASSERT_SUCCESS(aws_s3_tester_init(allocator, &tester));
+    struct aws_s3_tester_client_options client_options = {
+        .part_size = MB_TO_BYTES(5),
+        .tls_usage = AWS_S3_TLS_DISABLED,
+    };
+
+    struct aws_s3_client *client = NULL;
+    ASSERT_SUCCESS(aws_s3_tester_client_new(&tester, &client_options, &client));
+    struct aws_s3_client_vtable *patched_client_vtable = aws_s3_tester_patch_client_vtable(&tester, client, NULL);
+    patched_client_vtable->after_prepare_upload_part_finish_stub =
+        s_after_prepare_upload_part_finish_stub_retry_before_finish_sending;
+    struct aws_s3_file_io_options fio_opts = {
+        .should_stream = true,
+        .direct_io = true,
+    };
+    struct aws_byte_cursor object_path = aws_byte_cursor_from_c_str("/default");
+
+    /* retry with streaming upload. */
+    struct aws_s3_tester_meta_request_options put_options = {
+        .allocator = allocator,
+        .meta_request_type = AWS_S3_META_REQUEST_TYPE_PUT_OBJECT,
+        .client = client,
+        .checksum_algorithm = AWS_SCA_CRC32,
+        .validate_get_response_checksum = false,
+        .fio_opts = &fio_opts,
+        .put_options =
+            {
+                .file_on_disk = true,
+                .object_size_mb = 10,
+                .object_path_override = object_path,
+            },
+        .mock_server = true,
+    };
+
+    struct aws_s3_meta_request_test_results meta_request_test_results;
+    aws_s3_meta_request_test_results_init(&meta_request_test_results, allocator);
+
+    ASSERT_SUCCESS(aws_s3_tester_send_meta_request_with_options(&tester, &put_options, &meta_request_test_results));
+
+    ASSERT_INT_EQUALS(meta_request_test_results.upload_review.part_count, 2);
+    /* Note: the data we currently generate is always the same,
+     * so make sure that retry does not mangle the data by checking the checksum value */
+    ASSERT_STR_EQUALS("7/xUXw==", aws_string_c_str(meta_request_test_results.upload_review.part_checksums_array[0]));
+    ASSERT_STR_EQUALS("PCOjcw==", aws_string_c_str(meta_request_test_results.upload_review.part_checksums_array[1]));
+    aws_s3_meta_request_test_results_clean_up(&meta_request_test_results);
+    aws_s3_client_release(client);
+    aws_s3_tester_clean_up(&tester);
+
+    return AWS_OP_SUCCESS;
+}
+
 TEST_CASE(multipart_download_checksum_with_retry_mock_server) {
     (void)ctx;
     /**
